import os
import os.path
import sys
import typing
import urllib.parse
import urllib.request
import validators

from . import log
from . import util



# constants

_ENCODING = "UTF-8"
_SUPPORTED_SCHEMES = {"http", "https", "ftp"}
_TTY_DEVICE_FILE = "/dev/tty" # POSIX compliant
_PATH_TYPE = typing.Union[str, None]



# functions

def _is_uri(uri):
  return validators.url(uri)


def _read_from_stdin():
  try:
    text = sys.stdin.read()
  except KeyboardInterrupt:
    log.error("Nothing could be read from stdin.")

  os.close(sys.stdin.fileno())
  sys.stdin = open(_TTY_DEVICE_FILE)

  return text


def _read_local_file(path):
  try:
    with open(path, "rb") as f:
      return f.read().decode(_ENCODING, "replace")
  except (FileNotFoundError, PermissionError) as exception:
    log.error(exception)


def _read_remote_file(uri):
  if urllib.parse.urlparse(uri).scheme not in _SUPPORTED_SCHEMES:
    log.error("Invalid scheme of URI is detected. "
              "(supported schemes: {})"
              .format(util.sequence_to_string(_SUPPORTED_SCHEMES)))

  log.message("Loading a page...")
  try:
    with urllib.request.urlopen(uri) as response:
      return response.read().decode(_ENCODING, "replace")
  except urllib.error.URLError as exception:
    log.error(exception)


def _uri_to_filename(uri):
  return os.path.basename(urllib.parse.urlparse(uri).path)


def path_to_filename(path: _PATH_TYPE):
  if path is None:
    return None
  elif _is_uri(path):
    return _uri_to_filename(path)
  return os.path.basename(path)


def path_to_text(path: _PATH_TYPE):
  if path is None:
    return _read_from_stdin()
  elif _is_uri(path):
    return _read_remote_file(path)
  return _read_local_file(path)



#!/usr/bin/env python

import sys

from . import consolekit as ck
from .get_args import get_args
from .path_to_x import path_to_filename, path_to_text
from . import pygments_util
from . import shakyo
from . import text_to_lines
from . import log



def get_example_lines(example_path,
                      example_text,
                      console,
                      *,
                      lexer_name,
                      style_name,
                      colorize,
                      decorate):
  return text_to_lines.text_to_lines(
      example_text,
      console,
      lexer=pygments_util.guess_lexer(
          lexer_name=lexer_name,
          filename=path_to_filename(example_path),
          text=example_text),
      style_name=style_name,
      colorize=colorize,
      decorate=decorate)


def main():
  args = get_args()

  if not sys.stdout.isatty(): log.error("stdout is not a tty.")

  example_text = path_to_text(args.example_path)

  with ck.Console(asciize=args.asciize,
                  spaces_per_tab=args.spaces_per_tab,
                  background_rgb=args.background_rgb) as console:
    shakyo.Shakyo(console,
                  get_example_lines(args.example_path,
                                    example_text,
                                    console,
                                    lexer_name=args.lexer_name,
                                    style_name=args.style_name,
                                    colorize=args.colorize,
                                    decorate=args.decorate)).do()


if __name__ == "__main__":
  main()



import pygments.formatter
import pygments.styles

from . import consolekit as ck
from . import util



# classes

class _AttrTable:
  def __init__(self, console, style="default", colorize=True, decorate=True):
    self._token_type_to_attr = {}
    for token_type, properties in self._style_to_token_properties(style):
      attr = console.decoration_attrs.normal
      if colorize and properties["color"]:
        attr |= console.color_attrs.get_best_match(
                util.interpret_string_rgb(properties["color"]))
      if decorate and properties["bold"]:
        attr |= console.decoration_attrs.bold
      if decorate and properties["underline"]:
        attr |= console.decoration_attrs.underline
      self._token_type_to_attr[token_type] = attr

  def __getitem__(self, token_type):
    while token_type not in self._token_type_to_attr:
      if token_type == token_type.parent:
        raise KeyError("Token type, {} is not found in attribute table."
                       .format(token_type))
      token_type = token_type.parent
    return self._token_type_to_attr[token_type]

  @staticmethod
  def _style_to_token_properties(style):
    return pygments.formatter.Formatter(style=style).style



# functions

def text_to_lines(text,
                  console,
                  lexer,
                  style_name="default",
                  colorize=True,
                  decorate=True):
  style = pygments.styles.get_style_by_name(style_name)
  attr_table = _AttrTable(console,
                          style=style,
                          colorize=colorize,
                          decorate=decorate)
  yield from _tokens_to_lines(lexer.get_tokens(_strip_text(text)), attr_table)


def _tokens_to_lines(tokens, attr_table):
  line = ck.Line()
  for token_type, string in tokens:
    for char in string:
      if char == '\n':
        yield line
        line = ck.Line()
      elif ck.is_printable_char(char):
        line += ck.Character(char, attr_table[token_type])

  # if there is no newline character at the end of the last line
  if len(line) > 0:
    yield line


def _strip_text(text):
  return '\n'.join(line.rstrip() for line in text.split('\n'))



import text_unidecode
import unicodedata

from . import character



class Line:
  """
  A class whose instances represent one-line text on a console.
  Line instances are immutable.
  """

  _SPACES_PER_TAB = 4
  _ASCIIZE = False

  def __init__(self, *chars):
    assert all(isinstance(char, character.Character) for char in chars)
    self._chars = chars

  def __len__(self):
    return len(self._chars)

  def __eq__(self, line):
    if line is None: return False
    assert isinstance(line, Line)

    if len(self) != len(line): return False
    for my_char, your_char in zip(self, line):
      if my_char != your_char: return False
    return True

  def __iter__(self):
    for char in self._chars:
      yield char

  def __add__(self, char_or_line):
    if isinstance(char_or_line, character.Character):
      char = char_or_line
      return Line(*self, char)
    assert isinstance(char_or_line, Line)
    line = char_or_line
    return Line(*self, *line)

  def __radd__(self, char):
    assert isinstance(char, character.Character)
    return Line(char, *self)

  def __getitem__(self, key):
    if isinstance(key, int):
      return self._chars[key]
    assert isinstance(key, slice)
    return Line(*self._chars[key])

  @property
  def normalized(self):
    return Line(*self._normalized_chars)

  @property
  def width(self):
    return sum(char.width for char in self.normalized)

  @property
  def _normalized_chars(self):
    position = 0
    for char in self._chars:
      if char == '\t':
        boundary = self._next_tab_boundary(position)
        while position != boundary:
          position += 1
          yield character.Character(' ', char.attr)
      else:
        for normalized_char in self._normalize_char(char):
          position += normalized_char.width
          yield normalized_char

  @classmethod
  def _next_tab_boundary(cls, position):
    return (position // cls._SPACES_PER_TAB + 1) * cls._SPACES_PER_TAB

  @classmethod
  def _normalize_char(cls, char):
    if cls._ASCIIZE:
      return [character.Character(string_char, char.attr)
              for string_char in text_unidecode.unidecode(str(char))]
    return [character.Character(string_char, char.attr)
            if not unicodedata.category(string_char).startswith("Z")
            else character.Character(' ', char.attr)
            for string_char in unicodedata.normalize("NFC", str(char))]



import pygments
import pygments.lexers
import pygments.styles
import pygments.lexers.special



# constants

_LEXER_OPTIONS = {"stripall" : True}
_FALLBACK_LEXER = pygments.lexers.special.TextLexer(**_LEXER_OPTIONS)



# functions

def guess_lexer(lexer_name=None, filename=None, text=None):
  if lexer_name is not None:
    return pygments.lexers.get_lexer_by_name(lexer_name)
  return _guess_lexer_from_filename_and_text(filename, text)


def _guess_lexer_from_filename_and_text(filename, text):
  lexer = None
  if filename is not None:
    lexer = _guess_lexer_from_filename(filename)
  if text is not None and lexer is None:
    lexer = _guess_lexer_from_text(text)
  if lexer is None:
    lexer = _FALLBACK_LEXER
  return lexer


def _guess_lexer_from_filename(filename):
  try:
    return pygments.lexers.get_lexer_for_filename(filename, **_LEXER_OPTIONS)
  except pygments.util.ClassNotFound:
    return None


def _guess_lexer_from_text(text):
  try:
    return pygments.lexers.guess_lexer(text, **_LEXER_OPTIONS)
  except pygments.util.ClassNotFound:
    return None


def all_lexer_names():
  return {alias for _, aliases, _, _ in pygments.lexers.get_all_lexers()
          for alias in aliases}


def all_style_names():
  return pygments.styles.get_all_styles()



import os.path
import sys

from . import consolekit as ck



VERSION = "0.0.10"

COMMAND_NAME = os.path.basename(sys.argv[0])

DELETE_CHARS = ck.DELETE_CHARS | ck.BACKSPACE_CHARS
QUIT_CHARS = ck.ESCAPE_CHARS
CLEAR_CHAR = ck.ctrl('u')
SCROLL_DOWN_CHAR = ck.ctrl('n')
SCROLL_UP_CHAR = ck.ctrl('p')
PAGE_DOWN_CHAR = ck.ctrl('f')
PAGE_UP_CHAR = ck.ctrl('b')



import unicodedata

from . import attribute
from . import misc



class Character:
  def __init__(self, string, attr=attribute.DecorationAttribute.normal):
    assert misc.is_printable_char(string) and isinstance(attr, int)
    self._string = string
    self._attr = attr

  def __str__(self):
    return self._string

  def __eq__(self, you):
    return str(self) == str(you)

  @property
  def attr(self):
    return self._attr

  @property
  def width(self):
    return 2 if unicodedata.east_asian_width(self._string) in {"W", "F"} else 1



def interpret_string_rgb(string_rgb):
  assert len(string_rgb) == 6
  int_rgb = int(string_rgb, 16)
  return (int_rgb >> 16 & 0xff, int_rgb >> 8 & 0xff, int_rgb & 0xff)


def sequence_to_string(sequence):
  return ", ".join(sorted(str(elem) for elem in sequence))



import argparse

from . import consolekit as ck
from . import config
from . import log
from . import pygments_util
from . import util



# constants

_DEFAULT_ARGUMENT_HELP = " (default: %(default)s)"
_DESCRIPTION = "{} is a tool to learn about something just by typing it. " \
               "To play with it, just type letters on a cursor. " \
               "Type {} to scroll down and {} to scroll up one line, " \
               "{} to scroll down and {} to scroll up one page, " \
               "and Esc or ^[ to exit while running it." \
               .format(config.COMMAND_NAME,
                       ck.unctrl(config.SCROLL_DOWN_CHAR),
                       ck.unctrl(config.SCROLL_UP_CHAR),
                       ck.unctrl(config.PAGE_DOWN_CHAR),
                       ck.unctrl(config.PAGE_UP_CHAR))
_SHOW_LANGUAGES_OPTION = "--show-languages"
_SHOW_STYLES_OPTION = "--show-styles"



# functions

def get_args():
  arg_parser = argparse.ArgumentParser(description=_DESCRIPTION)

  arg_parser.add_argument("example_path", nargs='?', default=None,
                          help="file path or URI to an example")
  arg_parser.add_argument("-a", "--asciize",
                          dest="asciize", action="store_true",
                          help="enable asciization of unicode characters")
  arg_parser.add_argument("-b", "--background-color",
                          dest="background_rgb", default="000000",
                          help="tell {} the hexadecimal background color "
                               "of your terminal to avoid the same font color "
                               "as it"
                               .format(config.COMMAND_NAME)
                               + _DEFAULT_ARGUMENT_HELP)
  arg_parser.add_argument("-c", "--no-color",
                          dest="colorize", action="store_false",
                          help="disable colorization of text")
  arg_parser.add_argument("-d", "--no-decoration",
                          dest="decorate", action="store_false",
                          help="disable decoration of text")
  arg_parser.add_argument("-l", "--language", metavar="LANGUAGE",
                          dest="lexer_name", type=str, default=None,
                          help="specify a language of an example")
  arg_parser.add_argument(_SHOW_LANGUAGES_OPTION,
                          dest="show_languages", action="store_true",
                          help="show all lauguages available for examples")
  arg_parser.add_argument("-s", "--style",
                          dest="style_name", type=str, default="default",
                          help="specify a style name" + _DEFAULT_ARGUMENT_HELP)
  arg_parser.add_argument(_SHOW_STYLES_OPTION,
                          dest="show_styles", action="store_true",
                          help="show all available style names")
  arg_parser.add_argument("-t", "--spaces-per-tab",
                          dest="spaces_per_tab", type=int, default=4,
                          help="set number of spaces per tab"
                               + _DEFAULT_ARGUMENT_HELP)
  arg_parser.add_argument("-v", "--version",
                          dest="show_version", action="store_true",
                          help="show version information")

  args = arg_parser.parse_args()

  if args.show_version:
    print("version:", config.VERSION)
    exit()
  elif args.show_languages:
    _print_sequence(pygments_util.all_lexer_names())
    exit()
  elif args.show_styles:
    _print_sequence(pygments_util.all_style_names())
    exit()

  try:
    args.background_rgb = util.interpret_string_rgb(args.background_rgb)
  except (AssertionError, ValueError):
    log.error("\"{}\" is invalid as a hexadecimal RGB color."
              .format(args.background_rgb))

  _check_args(args)

  return args


def _check_args(args):
  if args.spaces_per_tab <= 0:
    log.error("Number of spaces per tab must be greater than 0.")
  elif args.lexer_name is not None \
       and args.lexer_name not in pygments_util.all_lexer_names():
    log.error("The language, \"{}\" is not available for examples. "
              "See `{} {}`."
              .format(args.lexer_name,
                      config.COMMAND_NAME,
                      _SHOW_LANGUAGES_OPTION))
  elif args.style_name not in pygments_util.all_style_names():
    log.error("The style, \"{}\" is not available. See `{} {}`."
              .format(args.style_name,
                      config.COMMAND_NAME,
                      _SHOW_STYLES_OPTION))


def _print_sequence(sequence):
  print(util.sequence_to_string(sequence))



#!/usr/bin/env python

import os
import re
import setuptools
import shutil
import sys

if not sys.version_info >= (3, 5):
  exit("Sorry, Python's version must be later than 3.5.")



# constants

COMMAND_NAME = "shakyo"
README_BASENAME = "README"
RST_EXT = "rst"
MD_EXT = "md"



# functions

def warn(*text):
  print("WARNING:", *text, file=sys.stderr)


def version():
  with open("shakyo/config.py") as f:
    return next(re.match(r"_*VERSION\s*=\s*\"((\d|\.)*)\"", line)
                for line in f.readlines()
                if line.startswith("VERSION")).group(1)


def read_text_file(filename):
  with open(os.path.join(os.path.dirname(__file__), filename)) as f:
    return f.read()


def readme_text():
  readme_rst = README_BASENAME + "." + RST_EXT
  readme_md = README_BASENAME + "." + MD_EXT

  try:
    import pypandoc
    with open(readme_rst, "w") as f:
      f.write(pypandoc.convert(readme_md, RST_EXT))
    return read_text_file(readme_rst)
  except ImportError as e:
    warn(e)
    if os.path.isfile(readme_rst):
      os.remove(readme_rst)
    shutil.copyfile(readme_md, README_BASENAME)
    return read_text_file(README_BASENAME)



# main routine

def main():
  setuptools.setup(
      name=COMMAND_NAME,
      version=version(),
      description="a tool to learn about something just by typing it",
      long_description=readme_text(),
      license="Public Domain",
      author="raviqqe",
      author_email="raviqqe@gmail.com",
      url="http://github.com/raviqqe/shakyo/",
      packages=["shakyo", "shakyo.consolekit"],
      entry_points={
        "console_scripts" : [COMMAND_NAME + "=shakyo.__main__:main"]
      },
      install_requires=["pygments", "text_unidecode", "validators"],
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Environment :: Console :: Curses",
          "Intended Audience :: Developers",
          "Intended Audience :: End Users/Desktop",
          "License :: Public Domain",
          "Operating System :: POSIX",
          "Programming Language :: Python :: 3.5",
          "Topic :: Education :: Computer Aided Instruction (CAI)",
          "Topic :: Games/Entertainment",
      ],
  )


if __name__ == "__main__":
  main()



import curses

from .character import Character
from .console import Console
from .line import Line
from .misc import ESCAPE_CHARS, DELETE_CHARS, BACKSPACE_CHARS, \
                  is_printable_char, ctrl, unctrl



import curses
import curses.ascii
import unicodedata
from curses.ascii import ctrl, unctrl



ESCAPE_CHARS = {chr(curses.ascii.ESC), curses.ascii.ctrl('[')}
DELETE_CHARS = {chr(curses.ascii.DEL), chr(curses.KEY_DC)}
BACKSPACE_CHARS = {chr(curses.ascii.BS), chr(curses.KEY_BACKSPACE)}


def is_printable_char(char):
  return char == '\t' or not unicodedata.category(char).startswith("C")



from . import consolekit as ck
from . import config



# classes

class Shakyo:
  CURSOR_WIDTH = 1

  def __init__(self, console, example_lines):
    self._console = console
    self._geometry = _Geometry(console)
    self._input_line = ck.Line()
    self._example_lines = _FoldedLines(example_lines,
                                       max_width=(console.screen_width - 1))
    if self._example_lines[0] is None:
      raise Exception("No line can be read from the example source.")

  def do(self):
    self._print_all_example_lines()

    while self._example_lines[0] is not None:
      self._update_input_line()
      char = self._console.get_char()

      if char in config.QUIT_CHARS:
        break
      elif char == config.CLEAR_CHAR:
        self._input_line = ck.Line()
      elif char in config.DELETE_CHARS:
        self._input_line = self._input_line[:-1]
      elif char == config.PAGE_DOWN_CHAR:
        self._input_line = ck.Line()
        self._page_down()
      elif char == config.PAGE_UP_CHAR:
        self._input_line = ck.Line()
        self._page_up()
      elif char == config.SCROLL_UP_CHAR:
        self._input_line = ck.Line()
        self._scroll_up()
      elif (char == '\n' and self._input_line.normalized
                             == self._example_lines[0].normalized) \
           or (char == config.SCROLL_DOWN_CHAR):
        self._input_line = ck.Line()
        self._scroll_down()
      elif ck.is_printable_char(char) \
           and (self._input_line + ck.Character(char)).width \
               + self.CURSOR_WIDTH <= self._console.screen_width:
        self._input_line += ck.Character(char,
                                         self._next_input_char_attr(char))

  def _update_input_line(self):
    self._console.print_line(self._geometry.y_input, self._example_lines[0])
    self._console.print_line(self._geometry.y_input,
                             self._input_line,
                             clear=False)

  def _scroll_down(self):
    self._example_lines.base_index += 1
    bottom_line_index = self._geometry.y_bottom - self._geometry.y_input
    if self._example_lines[bottom_line_index] is not None:
      self._console.scroll(self._example_lines[bottom_line_index])
    else:
      self._console.scroll()

  def _scroll_up(self):
    if self._example_lines[-1] is None: return
    self._example_lines.base_index -= 1
    top_line_index = 0 - self._geometry.y_input
    if self._example_lines[top_line_index] is not None:
      self._console.scroll(self._example_lines[top_line_index], direction="up")
    else:
      self._console.scroll(direction="up")

  def _page_down(self):
    for _ in range(self._console.screen_height):
      if self._example_lines[1] is None: break
      self._scroll_down()

  def _page_up(self):
    for _ in range(self._console.screen_height):
      if self._example_lines[-1] is None: break
      self._scroll_up()

  def _print_all_example_lines(self):
    for index in range(self._geometry.y_bottom - self._geometry.y_input + 1):
      if self._example_lines[index] is None: break
      self._console.print_line(self._geometry.y_input + index,
                               self._example_lines[index])

  def _next_input_char_attr(self, char):
    attr_correct = self._console.decoration_attrs.normal
    attr_wrong = self._console.decoration_attrs.reverse
    return (attr_correct if self._is_correct_char(char) else attr_wrong) \
           | self._next_example_char_attr

  @property
  def _next_example_char_attr(self):
    normalized_example_line = self._example_lines[0].normalized
    return normalized_example_line[min(len(self._input_line.normalized),
                                       len(normalized_example_line) - 1)].attr\
           if len(normalized_example_line) != 0 else \
           self._console.decoration_attrs.normal

  def _is_correct_char(self, char):
    normalized_input_line = self._input_line.normalized
    normalized_example_line = self._example_lines[0].normalized
    normalized_new_input_line = (self._input_line + ck.Character(char)) \
                                .normalized

    if len(normalized_new_input_line) > len(normalized_example_line):
      return False
    new_char_range = slice(len(normalized_input_line),
                           len(normalized_new_input_line))
    if normalized_new_input_line[new_char_range] \
       != normalized_example_line[new_char_range]:
      return False
    return True


class _Geometry:
  def __init__(self, console):
    self.y_input = (console.screen_height - 1) // 2
    self.y_bottom = console.screen_height - 1


class _FoldedLines:
  def __init__(self, raw_lines, max_width=79):
    assert max_width >= 2 # for double-width characters
    self._line_generator = self._fold_lines(raw_lines, max_width)
    self._lines = []
    self._base_index = 0

  def __getitem__(self, relative_index):
    assert isinstance(relative_index, int)

    index = self._base_index + relative_index

    for line in self._line_generator:
      self._lines.append(line)
      if index < len(self._lines):
        break

    return self._lines[index] if 0 <= index < len(self._lines) else None

  @property
  def base_index(self):
    return self._base_index

  @base_index.setter
  def base_index(self, base_index):
    assert isinstance(base_index, int)
    self._base_index = base_index

  @classmethod
  def _fold_lines(cls, lines, max_width):
    for line in lines:
      while line.width > max_width:
        new_line, line = cls._split_line(line, max_width)
        yield new_line
      yield line

  @staticmethod
  def _split_line(line, max_width):
    assert line.width > max_width

    # binary search for max index to construct a line of max width
    min_index = 0
    max_index = len(line)
    while min_index != max_index:
      middle_index = (max_index - min_index + 1) // 2 + min_index
      if line[:middle_index].width <= max_width:
        min_index = middle_index
      else:
        max_index = middle_index - 1

    assert line[:min_index].width <= max_width
    return line[:min_index], line[min_index:]



import curses
import enum



_CURSES_COLOR_SCALE = 1000



@enum.unique
class DecorationAttribute(enum.IntEnum):
  altcharset =  curses.A_ALTCHARSET
  blink =       curses.A_BLINK
  bold =        curses.A_BOLD
  dim =         curses.A_DIM
  normal =      curses.A_NORMAL
  reverse =     curses.A_REVERSE
  standout =    curses.A_STANDOUT
  underline =   curses.A_UNDERLINE


class ColorAttribute:
  _colors = []
  _rgb_to_attr = {}
  _background_rgb = None

  @classmethod
  def initialize(cls, background_rgb=(0, 0, 0)):
    if not curses.has_colors(): return
    cls._set_up_colors()
    cls._background_rgb = background_rgb

  @classmethod
  def get_best_match(cls, rgb):
    assert cls._is_valid_rgb(rgb)

    if not curses.has_colors():
      return curses.A_NORMAL

    if rgb in cls._rgb_to_attr:
      return cls._rgb_to_attr[rgb]

    attr = cls._find_best_match(rgb)
    cls._rgb_to_attr[rgb] = attr
    return attr

  @classmethod
  def _set_up_colors(cls):
    for color_index in range(curses.COLORS):
      pair_index = color_index + 1
      if pair_index >= curses.COLOR_PAIRS: break
      curses.init_pair(pair_index, color_index, -1)
      cls._colors.append((cls._get_rgb_by_color_index(color_index),
                          curses.color_pair(pair_index)))
    cls._colors.sort()

  @classmethod
  def _find_best_match(cls, new_rgb):
    assert cls._is_valid_rgb(new_rgb)
    matched_attr = curses.color_pair(0) # white on black
    min_distance = cls._color_distance((0, 0, 0), (255, 255, 255))
    for rgb, attr in cls._colors:
      if rgb == cls._background_rgb: continue
      distance = cls._color_distance(new_rgb, rgb)
      if distance < min_distance:
        min_distance = distance
        matched_attr = attr
    return matched_attr

  @classmethod
  def _color_distance(cls, rgb1, rgb2):
    assert cls._is_valid_rgb(rgb1) and cls._is_valid_rgb(rgb2)
    return sum((color1 - color2) ** 2 for color1, color2 in zip(rgb1, rgb2))

  @staticmethod
  def _is_valid_rgb(rgb):
    return len(rgb) == 3 and all(isinstance(color, int) and 0 <= color < 256
                                 for color in rgb)

  @staticmethod
  def _get_rgb_by_color_index(color_index):
    return tuple((color * 255) // _CURSES_COLOR_SCALE
                 for color in curses.color_content(color_index))



import curses
import unicodedata

from . import line as ln
from . import attribute



class Console:
  def __init__(self,
               asciize=False,
               spaces_per_tab=4,
               background_rgb=(0, 0, 0)):
    ln.Line._ASCIIZE = asciize
    ln.Line._SPACES_PER_TAB = spaces_per_tab
    self._background_rgb = background_rgb

  def _initialize_window(self):
    self._window = curses.initscr()
    curses.noecho()
    curses.cbreak()
    self._window.keypad(True)
    self._window.scrollok(True)
    self._window.clear()
    self._window.move(0, 0)
    self._window.refresh()

  def _initialize_colors(self):
    curses.start_color()
    curses.use_default_colors()
    attribute.ColorAttribute.initialize(background_rgb=self._background_rgb)

  def turn_on(self):
    self._initialize_window()
    self._initialize_colors()

  def turn_off(self):
    curses.nocbreak()
    curses.echo()
    curses.endwin()

  def __enter__(self):
    self.turn_on()
    return self

  def __exit__(self, exc_type, _exc_value, _traceback):
    self.turn_off()
    if exc_type == KeyboardInterrupt:
      return True

  @property
  def decoration_attrs(self):
    return attribute.DecorationAttribute

  @property
  def color_attrs(self):
    return attribute.ColorAttribute

  def _keep_position(method):
    def wrapper(self, *args, **keyword_args):
      position = self._window.getyx()
      result = method(self, *args, **keyword_args)
      self._window.move(*position)
      return result
    return wrapper

  @property
  def screen_height(self):
    return self._window.getmaxyx()[0]

  @property
  def screen_width(self):
    return self._window.getmaxyx()[1]

  def get_char(self) -> str:
    if hasattr(self._window, "get_wch"):
      char = self._window.get_wch()
    else:
      char = self._window.getch()

    if isinstance(char, int):
      return chr(char)
    assert isinstance(char, str)
    return char

  def print_line(self, y, line, clear=True):
    assert 0 <= y < self.screen_height
    assert isinstance(line, ln.Line)

    self._window.move(y, 0)
    if clear:
      self._window.clrtoeol()

    for char in line.normalized:
      assert not unicodedata.category(str(char)).startswith("C")
      if self._window.getyx()[1] + char.width > self.screen_width:
        break
      self._window.addstr(str(char), char.attr)

  @_keep_position
  def erase(self):
    self._window.erase()

  def refresh(self):
    self._window.refresh()

  @_keep_position
  def scroll(self, line=None, direction="down"):
    assert direction in {"up", "down"}
    assert isinstance(line, ln.Line) or line is None

    self._window.scroll(1 if direction == "down" else -1)

    if line is not None and direction == "down":
      self.print_line(self.screen_height - 1, line)
    elif line is not None and direction == "up":
      self.print_line(0, line)
